

name_list = ['itemid', 'cid', 'cross', 'gmv', 'ctr']
label_path = ['s3://jiayun.spark.data/wangqi/wide_deep/Feature/itemid/2019-09-03/part-00000-93ea99ba-05d6-4de9-a554-d2207e443431-c000.txt',
             's3://jiayun.spark.data/wangqi/wide_deep/Feature/cid/2019-09-03/part-00000-58e046e0-fe70-4673-899a-78fb71091d4c-c000.txt',
             's3://jiayun.spark.data/wangqi/wide_deep/Feature/idpair/2019-09-03/part-00000-9face73f-c5f1-4f6f-ae98-5370c41b2bac-c000.txt',
               3000, 3000]
value_path = ['s3://jiayun.spark.data/wangqi/wide_deep/v1/online_graph/txt/deep_layer_variable/pid_embedding.txt',
              's3://jiayun.spark.data/wangqi/wide_deep/v1/online_graph/txt/wide_layer_variable/cid_embedding.txt',
              's3://jiayun.spark.data/wangqi/wide_deep/v1/online_graph/txt/wide_layer_variable/click_cross_embedding.txt',
              's3://jiayun.spark.data/wangqi/wide_deep/v1/online_graph/txt/wide_layer_variable/gmv_embedding.txt',
              's3://jiayun.spark.data/wangqi/wide_deep/v1/online_graph/txt/wide_layer_variable/ctr_embedding.txt']



# hadoop fs -cp s3://jiayun.spark.data/wangqi/wide_deep/v1/online_graph/txt/wide_layer_variable/ctr_embedding.txt /user/hadoop
# hadoop fs -copyToLocal /user/hadoop/ctr_embedding.txt

import pyspark
from pyspark.sql.functions import *
from pyspark.sql import SparkSession
from pyspark.sql.functions import udf
from functools import reduce
from pyspark.sql.window import Window
from pyspark.sql.functions import row_number

spark = SparkSession.builder.appName('tmp2').getOrCreate()

out = []


def concat(key):
    @udf
    def wrapper(l):
        return key+'_'+l
    return wrapper

merge_list = []
for key, label_path_, value_path_ in zip(name_list[1:], label_path[1:], value_path[1:]):
    if isinstance(label_path_, str):
        label = spark.read.text(label_path_)
        new_row = spark.createDataFrame([['DEFAULT']])
        label = new_row.union(label).withColumn('label', concat(key)('_1')).drop('_1')
        label = label.withColumn('id', row_number().over(Window.orderBy(monotonically_increasing_id())) - 1)

        value = spark.read.text(value_path_).withColumnRenamed('_1', 'value')
        value = value.withColumn('id', row_number().over(Window.orderBy(monotonically_increasing_id())) - 1)

        merge = label.join(value, label.id==value.id, 'inner').drop('id')
        merge_list.append(merge)

out = reduce(lambda x, y: x.union(y), merge_list)
out.withColumn('merge', concat_ws(',', out['label'], out['value'])).select('merge')\
    .write.text('s3://jiayun.spark.data/wangqi/wide_deep/online/data/wide')


merge_list = []
for key, label_path_, value_path_ in zip(name_list[:1], label_path[:1], value_path[:1]):
    if isinstance(label_path_, str):
        label = spark.read.text(label_path_)
        new_row = spark.createDataFrame([['DEFAULT']])
        label = new_row.union(label).withColumn('label', concat(key)('_1')).drop('_1')
        label = label.withColumn('id', row_number().over(Window.orderBy(monotonically_increasing_id())) - 1)

        value = spark.read.text(value_path_).withColumnRenamed('_1', 'value')
        value = value.withColumn('id', row_number().over(Window.orderBy(monotonically_increasing_id())) - 1)

        merge = label.join(value, label.id==value.id, 'inner').drop('id')
        merge_list.append(merge)

out = reduce(lambda x, y: x.union(y), merge_list)
out.withColumn('merge', concat_ws(',', out['label'], out['value'])).select('merge')\
    .write.text('s3://jiayun.spark.data/wangqi/wide_deep/online/data/deep')



# Copyright 2017 Yahoo Inc.
# Licensed under the terms of the Apache 2.0 license.
# Please see LICENSE file in the project root for terms.
"""This module provides helper functions for the TensorFlow application.

Primarily, these functions help with:

* starting the TensorFlow ``tf.train.Server`` for the node (allocating GPUs as desired, and determining the node's role in the cluster).
* managing input/output data for *InputMode.SPARK*.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import nested_scopes
from __future__ import print_function

import getpass
import logging
import os
import time
from six.moves.queue import Empty
from . import marker


def hdfs_path(ctx, path):
  """Convenience function to create a Tensorflow-compatible absolute HDFS path from relative paths

  Args:
    :ctx: TFNodeContext containing the metadata specific to this node in the cluster.
    :path: path to convert

  Returns:
    An absolute path prefixed with the correct filesystem scheme.
  """
  #  All Hadoop-Compatible File System Schemes (as of Hadoop 3.0.x):
  HADOOP_SCHEMES = ['adl://',
                    'file://',
                    'hdfs://',
                    'oss://',
                    's3://',
                    's3a://',
                    's3n://',
                    'swift://',
                    'viewfs://',
                    'wasb://']
  if (any(path.startswith(scheme) for scheme in HADOOP_SCHEMES)):
    # absolute path w/ scheme, just return as-is
    return path
  elif path.startswith("/"):
    # absolute path w/o scheme, just prepend w/ defaultFS
    return ctx.defaultFS + path
  else:
    # relative path, prepend defaultFS + standard working dir
    if ctx.defaultFS.startswith("hdfs://") or ctx.defaultFS.startswith("viewfs://"):
      return "{0}/user/{1}/{2}".format(ctx.defaultFS, getpass.getuser(), path)
    elif ctx.defaultFS.startswith("file://"):
      return "{0}/{1}/{2}".format(ctx.defaultFS, ctx.working_dir[1:], path)
    else:
      logging.warn("Unknown scheme {0} with relative path: {1}".format(ctx.defaultFS, path))
      return "{0}/{1}".format(ctx.defaultFS, path)


def start_cluster_server(ctx, num_gpus=1, rdma=False):
  """Function that wraps the creation of TensorFlow ``tf.train.Server`` for a node in a distributed TensorFlow cluster.

  This is intended to be invoked from within the TF ``map_fun``, replacing explicit code to instantiate ``tf.train.ClusterSpec``
  and ``tf.train.Server`` objects.

  Args:
    :ctx: TFNodeContext containing the metadata specific to this node in the cluster.
    :num_gpu: number of GPUs desired
    :rdma: boolean indicating if RDMA 'iverbs' should be used for cluster communications.

  Returns:
    A tuple of (cluster_spec, server)
  """
  import tensorflow as tf
  if tf.__version__[0]=='1':
    tfserver = tf.train.Server
  else:
    tfserver = tf.distribute.Server
  from . import gpu_info

  logging.info("{0}: ======== {1}:{2} ========".format(ctx.worker_num, ctx.job_name, ctx.task_index))
  cluster_spec = ctx.cluster_spec
  logging.info("{0}: Cluster spec: {1}".format(ctx.worker_num, cluster_spec))

  if tf.test.is_built_with_cuda() and num_gpus > 0:
    # compute my index relative to other nodes placed on the same host (for GPU allocation)
    my_addr = cluster_spec[ctx.job_name][ctx.task_index]
    my_host = my_addr.split(':')[0]
    flattened = [v for sublist in cluster_spec.values() for v in sublist]
    local_peers = [p for p in flattened if p.startswith(my_host)]
    my_index = local_peers.index(my_addr)

    # GPU
    gpu_initialized = False
    retries = 3
    while not gpu_initialized and retries > 0:
      try:
        # override PS jobs to only reserve one GPU
        if ctx.job_name == 'ps':
          num_gpus = 1

        # Find a free gpu(s) to use
        gpus_to_use = gpu_info.get_gpus(num_gpus, my_index)
        gpu_prompt = "GPU" if num_gpus == 1 else "GPUs"
        logging.info("{0}: Using {1}: {2}".format(ctx.worker_num, gpu_prompt, gpus_to_use))

        # Set GPU device to use for TensorFlow
        os.environ['CUDA_VISIBLE_DEVICES'] = gpus_to_use

        # Create a cluster from the parameter server and worker hosts.
        cluster = tf.train.ClusterSpec(cluster_spec)

        # Create and start a server for the local task.
        if rdma:
          server = tfserver(cluster, ctx.job_name, ctx.task_index, protocol="grpc+verbs")
        else:
          server = tfserver(cluster, ctx.job_name, ctx.task_index)
        gpu_initialized = True
      except Exception as e:
        print(e)
        logging.error("{0}: Failed to allocate GPU, trying again...".format(ctx.worker_num))
        retries -= 1
        time.sleep(10)
    if not gpu_initialized:
      raise Exception("Failed to allocate GPU")
  else:
    # CPU
    os.environ['CUDA_VISIBLE_DEVICES'] = ''
    logging.info("{0}: Using CPU".format(ctx.worker_num))

    # Create a cluster from the parameter server and worker hosts.
    cluster = tf.train.ClusterSpec(cluster_spec)

    # Create and start a server for the local task.
    server = tfserver(cluster, ctx.job_name, ctx.task_index)

  return (cluster, server)


def next_batch(mgr, batch_size, qname='input'):
  """*DEPRECATED*. Use TFNode.DataFeed class instead."""
  raise Exception("DEPRECATED: Use TFNode.DataFeed class instead")


def export_saved_model(sess, export_dir, tag_set, signatures):
  """Convenience function to export a saved_model using provided arguments

  The caller specifies the saved_model signatures in a simplified python dictionary form, as follows::

    signatures = {
      'signature_def_key': {
        'inputs': { 'input_tensor_alias': input_tensor_name },
        'outputs': { 'output_tensor_alias': output_tensor_name },
        'method_name': 'method'
      }
    }

  And this function will generate the `signature_def_map` and export the saved_model.

  Args:
    :sess: a tf.Session instance
    :export_dir: path to save exported saved_model
    :tag_set: string tag_set to identify the exported graph
    :signatures: simplified dictionary representation of a TensorFlow signature_def_map

  Returns:
    A saved_model exported to disk at ``export_dir``.
  """
  import tensorflow as tf
  g = sess.graph
  g._unsafe_unfinalize()           # https://github.com/tensorflow/serving/issues/363
  builder = tf.compat.v1.saved_model.builder.SavedModelBuilder(export_dir)

  logging.info("===== signatures: {}".format(signatures))
  signature_def_map = {}
  for key, sig in signatures.items():
    signature_def_map[key] = tf.saved_model.build_signature_def(
        inputs={name: tf.compat.v1.saved_model.utils.build_tensor_info(tensor) for name, tensor in sig['inputs'].items()},
        outputs={name: tf.compat.v1.saved_model.utils.build_tensor_info(tensor) for name, tensor in sig['outputs'].items()},
        method_name=sig['method_name'] if 'method_name' in sig else key)
  logging.info("===== signature_def_map: {}".format(signature_def_map))
  builder.add_meta_graph_and_variables(
      sess,
      tag_set.split(','),
      signature_def_map=signature_def_map,
      clear_devices=True)
  g.finalize()
  builder.save()


def batch_results(mgr, results, qname='output'):
  """*DEPRECATED*. Use TFNode.DataFeed class instead."""
  raise Exception("DEPRECATED: Use TFNode.DataFeed class instead")


def terminate(mgr, qname='input'):
  """*DEPRECATED*. Use TFNode.DataFeed class instead."""
  raise Exception("DEPRECATED: Use TFNode.DataFeed class instead")


class DataFeed(object):
  """This class manages the *InputMode.SPARK* data feeding process from the perspective of the TensorFlow application.

  Args:
    :mgr: TFManager instance associated with this Python worker process.
    :train_mode: boolean indicating if the data feed is expecting an output response (e.g. inferencing).
    :qname_in: *INTERNAL_USE*
    :qname_out: *INTERNAL_USE*
    :input_mapping: *For Spark ML Pipelines only*.  Dictionary of input DataFrame columns to input TensorFlow tensors.
  """
  def __init__(self, mgr, train_mode=True, qname_in='input', qname_out='output', input_mapping=None):

    self.mgr = mgr
    self.train_mode = train_mode
    self.qname_in = qname_in
    self.qname_out = qname_out
    self.done_feeding = False
    self.input_tensors = [tensor for col, tensor in sorted(input_mapping.items())] if input_mapping is not None else None

  def next_batch(self, batch_size):
    """Gets a batch of items from the input RDD.

    If multiple tensors are provided per row in the input RDD, e.g. tuple of (tensor1, tensor2, ..., tensorN) and:

    * no ``input_mapping`` was provided to the DataFeed constructor, this will return an array of ``batch_size`` tuples,
      and the caller is responsible for separating the tensors.
    * an ``input_mapping`` was provided to the DataFeed constructor, this will return a dictionary of N tensors,
      with tensor names as keys and arrays of length ``batch_size`` as values.

    Note: if the end of the data is reached, this may return with fewer than ``batch_size`` items.

    Args:
      :batch_size: number of items to retrieve.

    Returns:
      A batch of items or a dictionary of tensors.
    """
    logging.debug("next_batch() invoked")
    queue = self.mgr.get_queue(self.qname_in)
    queue_out = self.mgr.get_queue(self.qname_out)
    tensors = [] if self.input_tensors is None else {tensor: [] for tensor in self.input_tensors}
    count = 0
    while count < batch_size: #  dead cur
      item = queue.get(block=True)  # queue_in
      if item is None:
        # End of Feed
        logging.info("next_batch() got None")
        queue.task_done()
        self.done_feeding = True
        break
      # elif type(item) is marker.EndBeforePartition:
      #   logging.info("next_batch() got EndBeforePartition")
      #   queue.task_done()
      #   if not self.train_mode and count >= 0:
      #     break
      elif type(item) is marker.EndPartition:
        # End of Partition
        logging.info("next_batch() got EndPartition")
        queue.task_done()
        queue_out.put(marker.EndPartition())
        if not self.train_mode:
          if count > 0:
            break
      else:
        # Normal item
        if self.input_tensors is None:
          tensors.append(item)
        else:
          for i in range(len(self.input_tensors)):
            tensors[self.input_tensors[i]].append(item[i])
        count += 1
        queue.task_done()
    logging.debug("next_batch() returning {0} items".format(count))
    return tensors

  def should_stop(self):
    """Check if the feed process was told to stop (by a call to ``terminate``)."""
    return self.done_feeding

  def batch_results(self, results):
    """Push a batch of output results to the Spark output RDD of ``TFCluster.inference()``.

    Note: this currently expects a one-to-one mapping of input to output data, so the length of the ``results`` array should match the length of
    the previously retrieved batch of input data.

    Args:
      :results: array of output data for the equivalent batch of input data.
    """
    logging.debug("batch_results() invoked")
    queue = self.mgr.get_queue(self.qname_out)
    for item in results:
      queue.put(item, block=True)
    logging.debug("batch_results() returning data")

  def terminate(self):
    """Terminate data feeding early.

    Since TensorFlow applications can often terminate on conditions unrelated to the training data (e.g. steps, accuracy, etc),
    this method signals the data feeding process to ignore any further incoming data.  Note that Spark itself does not have a mechanism
    to terminate an RDD operation early, so the extra partitions will still be sent to the executors (but will be ignored).  Because
    of this, you should size your input data accordingly to avoid excessive overhead.
    """
    logging.info("terminate() invoked")
    self.mgr.set('state', 'terminating')

    # drop remaining items in the queue
    queue = self.mgr.get_queue(self.qname_in)
    count = 0
    done = False
    while not done:
      try:
        queue.get(block=True, timeout=5)
        queue.task_done()
        count += 1
      except Empty:
        logging.info("dropped {0} items from queue".format(count))
        done = True



import tensorflow as tf
from . import model_dict
import logging

class modelFactory(object):
    @staticmethod
    def build(train_conf, model_conf, mode):
        model_type = train_conf['model_type']
        Model = model_dict[model_type](model_conf, mode)
        return Model





import tensorflow as tf
from .BaseModel import baseModel
try:
    from ..utils import layers
except:
    from utils import layers

# wide columns

class deepfm(baseModel):
    def __init__(self, model_conf, mode):
        super(deepfm, self).__init__(model_conf, mode)
        self.dropout_keep_fm = model_conf['dropout_keep_fm']
        self.dropout_keep_deep = model_conf['dropout_keep_deep']

        self.use_first_order = self.model_conf['first_order']
        self.dnn_hidden_units = self.model_conf['dnn_hidden_units']

    def _forward_model(self, is_training, *args, **kwargs):
        sparse_features, deep_features, dense_features = args
        with tf.compat.v1.variable_scope('trunk'):
            if self.use_first_order:
                deep_features, bias_features = tf.split(deep_features, [-1, 1], axis=2)
                bias_features = tf.squeeze(bias_features, axis=2)
                cate_bias_features, con_bias_features = tf.split(bias_features, [self.dims['cate_num'], self.dims['con_num']],
                                                                 axis=1)
            cate_deep_features, con_deep_features = tf.split(deep_features,
                                                             [self.dims['deep_num'] - self.dims['con_deep_num'],
                                                              self.dims['con_deep_num']], axis=1)
            with tf.compat.v1.variable_scope('fm'):
                if self.use_first_order:
                    y_first_order = tf.concat([cate_bias_features, con_bias_features], axis=1)
                    y_first_order = tf.compat.v1.layers.batch_normalization(y_first_order, training=is_training)
                    # ld = tf.compat.v1.get_default_graph().get_tensor_by_name("trunk/fm/batch_normalization/batchnorm/mul_2:0")
                    # ld = tf.tile(tf.expand_dims(ld, 0), tf.stack([tf.shape(y_first_order)[0], 1]))
                    y_first_order = tf.compat.v1.layers.dropout(y_first_order, 1-self.dropout_keep_fm, training=is_training)

                fm_features = tf.concat([con_deep_features, cate_deep_features], axis=1)
                square_sum = tf.reduce_sum(input_tensor=tf.square(fm_features), axis=1)
                sum_square = tf.square(tf.reduce_sum(input_tensor=fm_features, axis=1))
                y_second_order = 0.5 * tf.subtract(sum_square, square_sum)
                y_second_order = tf.compat.v1.layers.batch_normalization(y_second_order, training=is_training)
                y_second_order = tf.compat.v1.layers.dropout(y_second_order, 1-self.dropout_keep_fm, training=is_training)

            with tf.compat.v1.variable_scope('dnn'):
                deep_input = tf.concat([tf.reshape(cate_deep_features, [-1, (self.dims['deep_num']-self.dims['con_deep_num'])*self.dims['s_embed_size']]), dense_features], axis=1)
                deep_input = tf.compat.v1.layers.dropout(deep_input, 1 - self.dropout_keep_deep, training=is_training)
                for dim in self.dnn_hidden_units:
                    deep_input = layers.dense(deep_input, dim, tf.nn.relu, bn=True, training=is_training)
                    deep_input = tf.compat.v1.layers.dropout(deep_input, 1 - self.dropout_keep_deep, training=is_training)
            if self.use_first_order:
                merge = tf.concat([y_first_order, y_second_order, deep_input], axis=1)
            else:
                merge = tf.concat([y_second_order, deep_input], axis=1)
            out = layers.dense(merge, 1, bn=False, training=is_training)
        predictions = tf.nn.sigmoid(out, name=self.out_node_names[0])
        return predictions


    def get_loss(self, labels, predictions):
        labels = tf.cast(labels, tf.int32)
        l1_reg, l2_reg = self.model_conf['l1_reg'], self.model_conf['l2_reg']
        loss = tf.compat.v1.losses.log_loss(labels, predictions)
        T_vars = tf.compat.v1.trainable_variables()
        M_vars = [var for var in T_vars if var.name.startswith('trunk')]
        E_vars = list(set(T_vars)-set(M_vars))
        l1_loss, l2_loss = 0, 0
        if l2_reg > 0:
            for vars in E_vars:
                l2_loss += tf.keras.regularizers.l2(
                    0.5 * (l2_reg))(vars)
        if l1_reg > 0:
            for vars in M_vars:
                l1_loss += tf.keras.regularizers.l1(
                    l1_reg)(vars)
        loss = loss + l1_loss + l2_loss
        return loss


    # def get_precision(self, labels, predictions):
    #     labels = tf.cast(labels, tf.int32)
    #     labels = tf.expand_dims(labels, axis=1)
    #
    #     return tf.reduce_mean(
    #         input_tensor=tf.cast(
    #             tf.less(tf.abs(predictions - tf.cast(labels, tf.float32)), 0.5),
    #             tf.float32))

    def get_eval_metric_ops(self, labels, predictions):
        """Return a dict of the evaluation Ops.
        Args:
            labels (Tensor): Labels tensor for training and evaluation.
            predictions (Tensor): Predictions Tensor.
        Returns:
            Dict of metric results keyed by name.
        """
        labels = tf.cast(labels, tf.int32)
        auc = tf.compat.v1.metrics.auc(
            labels=labels,
            predictions=predictions,
            name='auc',
            curve='ROC',
        )

        return {
            'auc': auc
        }

    def get_predictions_out(self, features, predictions, pids, out_format):
        predictions_out = {id:tf.expand_dims(tf.identity(pids[id], name=id), axis=1) if id != 'out' else predictions
                           for id in out_format}
        return predictions_out



# Copyright 2017 Yahoo Inc.
# Licensed under the terms of the Apache 2.0 license.
# Please see LICENSE file in the project root for terms.

from __future__ import absolute_import
from __future__ import division
from __future__ import nested_scopes
from __future__ import print_function

import logging
import os
import socket
import subprocess
import errno
from socket import error as socket_error
from . import gpu_info


def single_node_env(num_gpus=1):
  """Setup environment variables for Hadoop compatibility and GPU allocation"""
  import tensorflow as tf
  # ensure expanded CLASSPATH w/o glob characters (required for Spark 2.1 + JNI)
  if 'HADOOP_PREFIX' in os.environ and 'TFOS_CLASSPATH_UPDATED' not in os.environ:
      classpath = os.environ['CLASSPATH']
      hadoop_path = os.path.join(os.environ['HADOOP_PREFIX'], 'bin', 'hadoop')
      hadoop_classpath = subprocess.check_output([hadoop_path, 'classpath', '--glob']).decode()
      os.environ['CLASSPATH'] = classpath + os.pathsep + hadoop_classpath
      os.environ['TFOS_CLASSPATH_UPDATED'] = '1'

  # reserve GPU, if requested
  if tf.test.is_built_with_cuda():
    gpus_to_use = gpu_info.get_gpus(num_gpus)
    logging.info("Using gpu(s): {0}".format(gpus_to_use))
    os.environ['CUDA_VISIBLE_DEVICES'] = gpus_to_use
  else:
    # CPU
    logging.info("Using CPU")
    os.environ['CUDA_VISIBLE_DEVICES'] = ''


def get_ip_address():
  """Simple utility to get host IP address."""
  try:
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip_address = s.getsockname()[0]
  except socket_error as sockerr:
    if sockerr.errno != errno.ENETUNREACH:
      raise sockerr
    ip_address = socket.gethostbyname(socket.getfqdn())
  finally:
    s.close()

  return ip_address


def find_in_path(path, file):
  """Find a file in a given path string."""
  for p in path.split(os.pathsep):
    candidate = os.path.join(p, file)
    if os.path.exists(candidate) and os.path.isfile(candidate):
      return candidate
  return False


def write_executor_id(num):
  """Write executor_id into a local file in the executor's current working directory"""
  with open("executor_id", "w") as f:
    f.write(str(num))


def read_executor_id():
  """Read worker id from a local file in the executor's current working directory"""
  with open("executor_id", "r") as f:
    return int(f.read())



'''
@author: xiongfei
@license: (C) Copyright 2013-2017, Node Supply Chain Manager Corporation Limited.
@contact: 386344277@qq.com
@file: preprocess.py
@time: 2018/8/28 下午12:14
@desc: shanghaijiaotong university
'''
from collections import defaultdict
import sys
import os
import random
tmp = 's3://jiayun.spark.data/wangqi/I2IRankData/logData'
# There are 13 integer features and 26 categorical features
continous_features = range(1, 14)
categorial_features = range(14, 40)

# Clip integer features. The clip point for each integer feature
# is derived from the 95% quantile of the total values in each feature
continous_clip = [20, 600, 100, 50, 64000, 500, 100, 50, 500, 10, 10, 10, 50]


class FeatureGenerator:
    def __init__(self, num_category_feature, num_continuous_feature):

        self.dicts = []
        self.num_category_feature = num_category_feature
        self.dicts_size = []
        for i in range(0, num_category_feature):
            self.dicts.append(defaultdict(int))

        self.num_continuous_feature = num_continuous_feature
        self.min = [sys.maxsize] * num_continuous_feature
        self.max = [-sys.maxsize] * num_continuous_feature

    def build(self, datafile, categorial_features, continous_features, cutoff=0):
        with open(datafile, 'r') as f:
            print('build for {}'.format(datafile))
            cnt = 1
            for line in f:
                if cnt % 100 == 0:
                    print('cnt: {}'.format(cnt))
                features = line.rstrip('\n').split('\t')
                for i in range(0, self.num_category_feature):
                    if features[categorial_features[i]] != '':
                        self.dicts[i][features[categorial_features[i]]] += 1

                for i in range(0, self.num_continuous_feature):
                    val = features[continous_features[i]]
                    if val != '':
                        val = int(val)
                        if val != '':
                            val = int(val)
                            if val > continous_clip[i]:
                                val = continous_clip[i]
                            self.min[i] = min(self.min[i], val)
                            self.max[i] = max(self.max[i], val)
                cnt += 1

            for i in range(0, self.num_category_feature):
                self.dicts[i] = filter(lambda x: x[1] >= cutoff, self.dicts[i].items())
                self.dicts[i] = sorted(self.dicts[i], key=lambda x: (-x[1], x[0]))
                vocabs, _ = list(zip(*self.dicts[i]))
                self.dicts[i] = dict(zip(vocabs, range(1, len(vocabs) + 1)))
                self.dicts[i]['<unk>'] = 0
                self.dicts_size.append(len(vocabs))
                print('categorial_feature {}: vocabs size {}'.format(i, len(vocabs)))

    def gen(self, datadir, outdir):
        random.seed(123)
        with open(os.path.join(outdir, 'train.txt'), 'w') as out_train:
            with open(os.path.join(outdir, 'valid.txt'), 'w') as out_valid:
                with open(os.path.join(datadir, 'train_sub.txt'), 'r') as f:
                    print('gen for train and valid')
                    cnt = 1
                    for line in f:
                        if cnt % 100000 == 0:
                            print('cnt: {}'.format(cnt))
                        features = line.rstrip('\n').split('\t')
                        continous_vals = []
                        for i in range(0, self.num_continuous_feature):
                            val = features[continous_features[i]]
                            if val == '':
                                val = 0.0
                            val = float(val)
                            val = (val - self.min[i]) / (self.max[i] - self.min[i])
                            continous_vals.append("{0:.6f}".format(val).rstrip('0')
                                                  .rstrip('.'))

                        categorial_vals = []
                        for i in range(0, self.num_category_feature):
                            key = features[categorial_features[i]]
                            if key not in self.dicts[i]:
                                val = self.dicts[i]['<unk>']
                            else:
                                val = self.dicts[i][key]
                            categorial_vals.append(str(val))

                        label = features[0]
                        if random.randint(0, 9999) % 10 != 0:
                            out_train.write('\t'.join(
                                [label] + continous_vals + categorial_vals) + '\n')
                        else:
                            out_valid.write('\t'.join(
                                [label] + continous_vals + categorial_vals) + '\n')
                        cnt += 1


def preprocess(datadir, outdir):
    dists = FeatureGenerator(len(categorial_features), len(continous_features))
    dists.build(os.path.join(datadir, 'train_sub.txt'), categorial_features, continous_features, cutoff=200)
    dists.gen(datadir, outdir)


if __name__ == "__main__":
    preprocess('/Users/xiongfei/Downloads/dac', './data')



import tensorflow as tf
from .BaseModel import baseModel
try:
    from ..utils import layers
except:
    from utils import layers


class deepfm(baseModel):
    def __init__(self, model_conf, mode):
        super(deepfm, self).__init__(model_conf, mode)
        self.dropout_keep_fm = model_conf['dropout_keep_fm']
        self.dropout_keep_deep = model_conf['dropout_keep_deep']

        self.use_first_order = self.model_conf['first_order']
        self.dnn_hidden_units = self.model_conf['dnn_hidden_units']

    def single_forward_model(self, is_training, deep_features, dense_features):
        with tf.compat.v1.variable_scope('trunk', reuse=True):
            if self.use_first_order:
                deep_features, bias_features = tf.split(deep_features, [-1, 1], axis=2)
                bias_features = tf.squeeze(bias_features, axis=2)
                cate_bias_features, con_bias_features = tf.split(bias_features, [self.dims['cate_num'], self.dims['con_num']],
                                                                 axis=1)
            cate_deep_features, con_deep_features = tf.split(deep_features,
                                                             [self.dims['deep_num'] - self.dims['con_deep_num'],
                                                              self.dims['con_deep_num']], axis=1)
            with tf.compat.v1.variable_scope('fm'):
                if self.use_first_order:
                    y_first_order = tf.concat([cate_bias_features, con_bias_features], axis=1)
                    y_first_order = tf.compat.v1.layers.batch_normalization(y_first_order, training=is_training)

                    y_first_order = tf.compat.v1.layers.dropout(y_first_order, 1-self.dropout_keep_fm, training=is_training)

                fm_features = tf.concat([con_deep_features, cate_deep_features], axis=1)
                square_sum = tf.reduce_sum(input_tensor=tf.square(fm_features), axis=1)
                sum_square = tf.square(tf.reduce_sum(input_tensor=fm_features, axis=1))
                y_second_order = 0.5 * tf.subtract(sum_square, square_sum)
                y_second_order = tf.compat.v1.layers.batch_normalization(y_second_order, training=is_training)
                y_second_order = tf.compat.v1.layers.dropout(y_second_order, 1-self.dropout_keep_fm, training=is_training)

            with tf.compat.v1.variable_scope('dnn'):
                deep_input = tf.concat([tf.reshape(cate_deep_features, [-1, (self.dims['deep_num']-self.dims['con_deep_num'])*self.dims['s_embed_size']]), dense_features], axis=1)
                deep_input = tf.compat.v1.layers.dropout(deep_input, 1 - self.dropout_keep_deep, training=is_training)
                for dim in self.dnn_hidden_units:
                    deep_input = layers.dense(deep_input, dim, tf.nn.relu, bn=True, training=is_training)
                    deep_input = tf.compat.v1.layers.dropout(deep_input, 1 - self.dropout_keep_deep, training=is_training)
            if self.use_first_order:
                merge = tf.concat([y_first_order, y_second_order, deep_input], axis=1)
            else:
                merge = tf.concat([y_second_order, deep_input], axis=1)
            out = layers.dense(merge, 1, bn=False, training=is_training)
        return out

    def _forward_model(self, is_training, *args, **kwargs):
        sparse_features, deep_features, dense_features = args
        deep_features_shape = deep_features.shape.as_list()
        dense_features_shape = dense_features.shape.as_list()
        if self.mode == 'train':
            deep_features = tf.reshape(deep_features, shape=[-1, 3, deep_features_shape[1]//3, deep_features_shape[2]])
            deep_features_split = tf.split(deep_features, 3, axis=1)
            deep_features_pair_p = tf.concat([deep_features_split[0], deep_features_split[1]], axis=1)
            deep_features_pair_n = tf.concat([deep_features_split[0], deep_features_split[2]], axis=1)

            dense_features = tf.reshape(dense_features, shape=[-1, 3, dense_features_shape[1]//3, dense_features_shape[2]])
            dense_features_split = tf.split(dense_features, 3, axis=1)
            dense_features_pair_p = tf.concat([dense_features_split[0], dense_features_split[1]], axis=1)
            dense_features_pair_n = tf.concat([dense_features_split[0], dense_features_split[2]], axis=1)

            out_p = self.single_forward_model(is_training, deep_features_pair_p, deep_features_pair_n)
            out_n = self.single_forward_model(is_training, dense_features_pair_p, dense_features_pair_n)
            pass
        else:
            out = self.single_forward_model(deep_features, dense_features)
            out = tf.identity(out, name=self.out_node_names[0])
            return out

    def get_eval_metric_ops(self, labels, predictions):
        """Return a dict of the evaluation Ops.
        Args:
            labels (Tensor): Labels tensor for training and evaluation.
            predictions (Tensor): Predictions Tensor.
        Returns:
            Dict of metric results keyed by name.
        """
        labels = tf.cast(labels, tf.int32)
        auc = tf.compat.v1.metrics.auc(
            labels=labels,
            predictions=predictions,
            name='auc',
            curve='ROC',
        )

        return {
            'auc': auc
        }

    def get_predictions_out(self, features, predictions, pids, out_format):
        predictions_out = {id:tf.expand_dims(tf.identity(pids[id], name=id), axis=1) if id != 'out' else predictions
                           for id in out_format}
        return predictions_out


import tensorflow as tf
from .BaseModel import baseModel
try:
    from ..utils import layers
except:
    from utils import layers

class xcross(baseModel):


    def _forward_model(self, sparse_features, deep_features, dense_features, is_training):
        with tf.compat.v1.variable_scope('trunk'):
            cate_deep_features, con_deep_features = tf.split(deep_features,
                                                             [self.dims['deep_num'] - self.dims['con_deep_num'],
                                                              self.dims['con_deep_num']], axis=1)
            x0 = tf.cate_deep_features, dense_features
            with tf.compat.v1.variable_scope('cross'):

                wide_out = layers.dense(sparse_features, 1)
            with tf.compat.v1.variable_scope('deep'):
                deep_input = tf.concat([tf.reshape(cate_deep_features, [-1, (self.dims['deep_num']-self.dims['con_deep_num'])*self.dims['s_embed_size']]), dense_features], axis=1)
                deep_input = tf.compat.v1.layers.dropout(deep_input, 1 - self.dropout_keep_deep, training=is_training)
                for dim in self.dnn_hidden_units:
                    deep_input = layers.dense(deep_input, dim, bn=True, activation=tf.nn.relu, training=is_training)
                    deep_input = tf.compat.v1.layers.dropout(deep_input, 1 - self.dropout_keep_deep, training=is_training)
                deep_out = layers.dense(deep_input, 1)
        out = wide_out + deep_out
        ld = wide_out
        return out, ld

    def forward(self, wide_features, deep_features, dense_features, deep_embedding_columns, load_conf, is_training):
        deep_input = tf.split(deep_features, num_or_size_splits=self.model_conf['field_num'], axis=1)
        deep_input = tf.concat([tf.expand_dims(feature, axis=1) for feature in deep_input], axis=1)
        cross_layer_size = self.model_conf['cross_layer_size']
        final_result = []
        final_len = 0
        embed_dim = self.model_conf['embed_dim']
        field_nums = [self.model_conf['field_num']]
        cin_layers = [deep_input]
        for idx, layer_size in enumerate(cross_layer_size):
            dot_result = tf.einsum('ipk,iqk->ipqk', cin_layers[0], cin_layers[-1])
            dot_result = tf.reshape(dot_result, shape=[-1, embed_dim, field_nums[0] * field_nums[-1]])
            filters = tf.get_variable(initializer=self.initializer_fn, name='f_' + str(idx),
                                      shape=[1, field_nums[-1] * field_nums[0], layer_size],
                                      dtype=tf.float32)

            # [batch_size, embed_dim, layers_size]
            curr_out = tf.nn.conv1d(dot_result, filters=filters, stride=1, padding='VALID')
            final_len += layer_size
            field_nums.append(layer_size)
            final_result.append(curr_out)
            curr_out = tf.transpose(curr_out, [0, 2, 1])
            cin_layers.append(curr_out)
        result = tf.concat(final_result, axis=2)
        result = tf.reduce_sum(result, axis=1)  # [batch_size, layer_size]

        input = tf.concat([wide_features, result, dense_features], axis=1)
        dnn_hidden_units = self.model_conf['dnn_hidden_units']
        for dim in dnn_hidden_units:
            input = tf.layers.dense(input, dim, activation=self.activation_fn,
                                    kernel_initializer=self.initializer_fn)
        input = tf.layers.dense(input, 1, activation=None, kernel_initializer=self.initializer_fn)
        return input